import sys
import types

# Python 3.x or 2.x
from exceptions import PrimaryKeyInInsertValues, PrimaryKeyMissingInInsertValues

PY3 = sys.version_info[0] == 3 or False
ITEM_LIMIT = 999  # How many items to cache before forcing an executemany.


class ItemCache(object):
    """
    Holds item waiting to be inserted or updated.
    """

    def __init__(self, model_context, operations, value_parsers={}, cache_limit=ITEM_LIMIT):
        """

        """
        self.value_parsers = value_parsers
        self.model_context = model_context
        self.operations = operations
        self.cache_limit = cache_limit
        self.reset()

    def flush(self):
        """

        """
        self.operations.execute_sql(self)

    def reset(self):
        """

        """
        self.item_counter = 0
        self.insert_items = []
        self.delete_items = []
        self.bulk_updates = {}
        self.update_items = []
        self.bulk_updates = {}
        for field in self.model_context.fields:
            self.bulk_updates[field] = {}

    def is_full(self):
        """

        """
        return self.item_counter >= self.cache_limit

    def _on_add(self):
        """

        """
        if self.is_full():
            self.operations.execute_sql(self)
            self.reset()

    def parse_values(self, values):
        """
        Executes any values parsers found in model.
        """
        for parser in self.value_parsers:
            values = parser(values)
            if not values:
                return
        return values

    def update(self, values):
        """
        Adds a set of values to execute as update using cursor.executemany.
        """
        if not self.model_context.pk in values:
            raise PrimaryKeyMissingInInsertValues(self.model_context.pk)

        values = self.parse_values(values)
        if not values:
            return

        self.update_items.append(values)
        self.item_counter += 1
        self._on_add()

    def bulk_update(self, values):
        """
        Adds a set of values to use for alternative bulk updates, using Model.objects.filter(id__in=...).update().
        See the README.md for more details.
        """
        pk = values.get(self.model_context.pk, None)
        if not pk:
            raise PrimaryKeyMissingInInsertValues("Missing primary key. Required for call to prepare.")

        values = self.parse_values(values)
        if not values:
            return

        del values[self.model_context.pk]
        for k, v in values.items():
            self.bulk_updates.setdefault(k, {}).setdefault(v, []).append(pk)
        self.item_counter += 1
        self._on_add()

    def insert(self, values):
        """
        Adds a dictionary with values to insert/update
        """
        if self.model_context.pk in values and self.model_context.pk_is_auto_field:
            raise PrimaryKeyInInsertValues('Primary key in insert values: %s' % values)

        final_values = {}
        for k, v in self.model_context.default_values.items():
            if callable(v):
                final_values[k] = v()
            else:
                final_values[k] = v

        final_values.update(values)

        final_values = self.parse_values(final_values)
        if not final_values:
            return

        self.insert_items.append(final_values)
        self.item_counter += 1
        self._on_add()

    def delete(self, pk):
        """
        Adds a primary key to the deletion queue.
        """
        if not PY3:
            assert types.IntType == type(pk), "pk argument must be integer."
        self.delete_items.append(int(pk))
        self.item_counter += 1
        self._on_add()



# -*- coding: latin-1 -*-
import random
import unittest
import time

import configure_app
configure_app.configure('tests')

from db_context import DbContext
from item_cache import ItemCache
from model_context import ModelContext
from context import DelayedContextFrom
from exceptions import PrimaryKeyInInsertValues, PrimaryKeyMissingInInsertValues

from django.test.utils import setup_test_environment, teardown_test_environment
from django.db import connection

from tests.models import foo, foobar, Child, ModelWithSQLKeywordAsField, WiktionaryPage, Photo, ModelWithValidation


class TestOperations(object):
    def execute_sql(self, item_cache):
        pass


class test_model_context(unittest.TestCase):

    def setUp(self):
        setup_test_environment()

    def tearDown(self):
        teardown_test_environment()

    def test_model_context1(self):
        db_context = DbContext(connection)
        m = ModelContext(foo, db_context)


class testMassiviu(unittest.TestCase):

    def setUp(self):
        setup_test_environment()

    def tearDown(self):
        teardown_test_environment()
        foo.objects.all().delete()
        foobar.objects.all().delete()

    def test_value_validator(self):
        def value_validator(values):
            if 'name' in values and len(values.get('name')) > 20:
                values['name'] = values['name'][:20]
            return values

        with DelayedContextFrom(foo, connection, value_validator) as cntx:
            cntx.insert({'name': 'Thomas'*50, 'age': 36, 'sex': 'M'})
        obj = foo.objects.first()
        self.assertEqual(len(obj.name), 20)

    def test_ItemCache(self):
        db_context = DbContext(connection)
        m = ModelContext(foo, db_context)
        x = ItemCache(m, TestOperations())
        for i in range(998):
            x.insert({'name': 'John Doe #%s' % i})
        self.assertEqual(x.item_counter, 998)
        for i in range(5):
            x.insert({'name': 'John Doe #%s' % i})
        self.assertEqual(x.item_counter, 4)

    def test_update(self):
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'name': 'John Doe #1'})
        self.assertEqual(foo.objects.all().count(), 1)
        id = foo.objects.first().id
        with DelayedContextFrom(foo) as cntx:
            cntx.update({'name': 'Thomas Weholt', 'id': id})
        self.assertEqual(foo.objects.get(id=id).name, 'Thomas Weholt')

    def test_values_factory_for_model(self):
        db_context = DbContext(connection)
        model_cnxt = ModelContext(foo, db_context)
        self.assertTrue(model_cnxt.default_values['sex'] == "M")
        self.assertTrue(model_cnxt.default_values['age'] == 20)
        self.assertTrue(model_cnxt.default_values.get('id', None) == None)

    def test_add_item(self):
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'name': 'Thomas', 'age': 36, 'sex': 'M'})
        self.assertTrue(foo.objects.all().count() == 1)

    def test_delete_item(self):
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'name': 'Thomas', 'age': 36, 'sex': 'M'})
        self.assertTrue(foo.objects.all().count() == 1)
        with DelayedContextFrom(foo) as cntx:
            cntx.delete(foo.objects.all()[0].id)
        self.assertTrue(foo.objects.all().count() == 0)

    def test_add_itemUsingModelDelayedExecutorUsingDefaults(self):
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'name': 'Thomas'})
        foo_record = foo.objects.all()[0]
        self.assertTrue(foo_record.age == 20)
        self.assertTrue(foo_record.sex == "M")

    def test_add_itemUsingCallableDefaultValue(self):
        with DelayedContextFrom(foobar, connection) as cntx:
            cntx.insert({'number': 1})
        time.sleep(0.5)
        with DelayedContextFrom(foobar, connection) as cntx:
            cntx.insert({'number': 2})
        item1 = foobar.objects.all()[0]
        item2 = foobar.objects.all()[1]
        self.assertTrue(item1.dt != item2.dt)

    def insertPersons(self, person_count):
        with DelayedContextFrom(foo) as cntx:
            for i in range(0, person_count):
                cntx.insert({'name': 'Person%s' % i, 'age': i})

    def test_reset_for_model(self):
        with DelayedContextFrom(foo) as cntx:
            for i in range(0, 10):
                cntx.insert({'name': 'Person%s' % i, 'age': i})
            cntx.reset()
        self.assertTrue(foo.objects.all().count() == 0)

    #
    #         # def test_sql_injection(self):
    #         # thanks to Cal Leeming @ Simplicity Media Ltd
    #         # """check out 'sqlmap',  http://sqlmap.sourceforge.net/doc/README.html#ss5.5
    #         # Here is a bunch of common injection techniques:
    #         # http://www.unixwiz.net/techtips/sql-injection.html"""
    #         # Write some tests here ...
    #
    def test_escaping_of_quotation_chars(self):
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'name': '"pony"'})
            cntx.insert({'name': "'pony'"})
            cntx.insert({'name': "`pony`"})
        self.assertTrue(foo.objects.all().count() == 3)

    def test_nonexisting_fields(self):
        # TODO : write some validation code and optionally crash on invalid data?
        with DelayedContextFrom(foo) as cntx:
            cntx.insert({'nosuchfield': 'Foo', 'anotherfield': 1, 'name': 'footest', 'age': 30})
        self.assertTrue(foo.objects.all().count() == 1)

    def test_update_only_update_specified_fields(self):
        params = {'name': 'Thomas', 'age': 36, 'sex': 'M'}
        with DelayedContextFrom(foo) as cntx:
            cntx.insert(params)
        object_id = foo.objects.first().id
        with DelayedContextFrom(foo) as cntx:
            cntx.update({'name': 'Thomas Weholt', 'id': object_id})

        updated_foo = foo.objects.all()[0]
        for k, v in {'id': object_id, 'name': 'Thomas Weholt', 'age': 36, 'sex': 'M'}.items():
            self.assertTrue(hasattr(updated_foo, k))
            self.assertTrue(getattr(updated_foo, k) == v)

    def test_bulk_update(self):
        params = {'name': 'Thomas', 'age': 36, 'sex': 'M'}
        with DelayedContextFrom(foo) as cntx:
            cntx.insert(params)
        object_id = foo.objects.first().id
        with DelayedContextFrom(foo) as cntx:
            cntx.bulk_update({'name': 'Thomas Weholt', 'id': object_id})

        updated_foo = foo.objects.first()
        for k, v in {'id': object_id, 'name': 'Thomas Weholt', 'age': 36, 'sex': 'M'}.items():
            self.assertTrue(hasattr(updated_foo, k))
            self.assertTrue(getattr(updated_foo, k) == v)

    def test_update_specific_fields(self):
        self.insertPersons(1)
        foo_id = foo.objects.first().id
        with DelayedContextFrom(foo) as cntx:
            cntx.update({'id': foo_id, 'name': 'Jo'})
        self.assertTrue(foo.objects.get(id=foo_id).name == 'Jo')

    def test_with_statement2(self):
        s = "was his name o'"
        self.insertPersons(10)
        with DelayedContextFrom(foo) as cntx:
            for item in foo.objects.all().values():
                cntx.update({'id': item.get('id'), 'name': "%s %s" % (item['name'], s)})
        self.assertTrue(s in foo.objects.all()[0].name)

    def test_implementing_subclass(self):
        with DelayedContextFrom(Child, connection) as cntx:
            for i in range(100):
                cntx.insert({'name': 'Whatever', 'age': random.randint(1, 20), })

    def test_sql_keyword_escaping(self):
        # "key" is a reserved word in SQL; for successful inserts/updates,
        # it must be escaped by Massiviu
        with DelayedContextFrom(ModelWithSQLKeywordAsField, connection) as cntx:
            cntx.insert(dict(key=1, where='select', update='*;'))
        record = ModelWithSQLKeywordAsField.objects.get()
        with DelayedContextFrom(ModelWithSQLKeywordAsField, connection) as cntx:
            cntx.update(dict(update=';update where', id=record.pk))

    def test_pk_field_is_autofield(self):
        with DelayedContextFrom(foo) as cntx:
            # Django's default PK
            self.assertEqual('id', cntx.model_context.pk)
            # The PK is an AutoField
            self.assertTrue(cntx.model_context.pk_is_auto_field)
            # Raise "PrimaryKeyInInsertValues" as expected
            self.assertRaises(PrimaryKeyInInsertValues, cntx.insert, {'id': 1, 'name': u"John Doe"})

    def test_pk_field_is_not_autofield(self):
        with DelayedContextFrom(WiktionaryPage, connection) as cntx:
            # The PK is still recognized
            self.assertEqual('id', cntx.model_context.pk)
            # The PK is not the AutoField generated by Django
            self.assertFalse(cntx.model_context.pk_is_auto_field)
            # Don't raise "PrimaryKeyInInsertValues" even if given
            cntx.insert({'id': 1234, 'title': u"django"})
        # Django didn't set the pk itself, else it would be "1"
        page = WiktionaryPage.objects.get(pk=1234)
        self.assertEqual(u"django", page.title)
        # Oops! didn't capitalize a proper name
        with DelayedContextFrom(WiktionaryPage, connection) as cntx:
            # Forgot to give the PK
            self.assertRaises(PrimaryKeyMissingInInsertValues, cntx.update, {'title': u"Django"})
            cntx.update({'id': 1234, 'title': u"Django"})
        page = WiktionaryPage.objects.get(pk=1234)
        self.assertEqual(u"Django", page.title)


class PerformanceTestMassiviu(unittest.TestCase):

    def setUp(self):
        setup_test_environment()

    def tearDown(self):
        foo.objects.all().delete()
        teardown_test_environment()

    def insertPersons(self, person_count):
        with DelayedContextFrom(foo) as cntx:
            for i in range(0, person_count):
                cntx.insert({'name': 'Person%s' % i, 'age': i})

    def test_SpeedReport_LotsOfInsertsUsingMassiviu(self):
        start_Massiviu = time.time()
        self.insertPersons(5000)
        stop_Massiviu = time.time()
        print("Using Massiviu took %s seconds." % (stop_Massiviu - start_Massiviu))

    def test_SpeedReport_LotsOfInsertsUsingOrm(self):
        start_orm = time.time()
        for i in range(0, 5000):
            foo(name='Person%s' % i, age=i).save()
        stop_orm = time.time()
        print("Using orm took %s seconds." % (stop_orm - start_orm))

    def test_SpeedReport_lotsOfGetAndUpdateUsingMassiviu(self):
        self.insertPersons(10000)
        start_Massiviu = time.time()
        with DelayedContextFrom(foo) as cntx:
            for item in foo.objects.all().values():
                item['name'] = "%s Doe" % item['name']
                cntx.update(item)
        stop_Massiviu = time.time()
        print("Using Massiviu took %s seconds." % (stop_Massiviu - start_Massiviu))

    def test_SpeedReport_lotsOfGetAndUpdateUsingOrm(self):
        self.insertPersons(10000)
        start_Massiviu = time.time()
        for item in foo.objects.all():
            item.name = "%s Doe" % item.name
            item.save()
        stop_Massiviu = time.time()
        print("Using orm took %s seconds." % (stop_Massiviu - start_Massiviu))

    def test_SpeedReportForUpdatesUsingMassiviu(self):
        self.insertPersons(5000)
        start_Massiviu = time.time()
        with DelayedContextFrom(foo) as cntx:
            for item in foo.objects.all().values():
                item['age'] += 10
                cntx.update(item)
        stop_Massiviu = time.time()
        print("Using Massiviu took %s seconds." % (stop_Massiviu - start_Massiviu))

    def test_SpeedReportForUpdatesUsingOrm(self):
        self.insertPersons(5000)
        start_orm = time.time()
        for item in foo.objects.all():
            item.name = "Person_%s" % item.id
            item.age += 10
            item.save()
        stop_orm = time.time()
        print("Using orm took %s seconds." % (stop_orm - start_orm))

    def test_deletion_queue(self):
        self.insertPersons(100)
        start = time.time()
        with DelayedContextFrom(foo) as cntx:
            for item in foo.objects.all().values():
                cntx.delete(item.get('id'))
        self.assertTrue(foo.objects.all().count() == 0)
        print("Took %s seconds to delete 10000 persons using Massiviu." % (time.time() - start))
        start = time.time()
        self.insertPersons(100)
        for p in foo.objects.all():
            p.delete()
        print("Took %s seconds to delete 10000 persons using orm." % (time.time() - start))

    def test_bulk_updates(self):
        camera_models = ('Nikon', 'Canon', 'Fujifilm', 'Panasonic', 'Sony', 'Leica', 'Pentax')
        with DelayedContextFrom(Photo, connection) as cntx:
            for i in range(0, 1000):
                cntx.insert({'filename': 'photo%s.jpg' % i, 'camera_model': 'Not set'})

        photo_to_camera_model_mapping = {}
        for i in range(0, 1000):
            photo_to_camera_model_mapping[i] = camera_models[random.randint(0, len(camera_models) - 1)]
        photo_to_rating_mapping = {}
        for i in range(0, 1000):
            photo_to_rating_mapping[i] = random.randint(0, 6)

        pks = [p.id for p in Photo.objects.all()]
        start = time.time()
        for i in pks:
            p = Photo.objects.get(id=i)
            p.camera_model = photo_to_camera_model_mapping.get(i, 'Not set')
            p.rating = photo_to_rating_mapping.get(i, 0)
            p.save()
        print("Took %s seconds to update 999 photos using orm." % (time.time() - start))
        Photo.objects.all().update(camera_model='Not set', rating=0)
        start = time.time()
        with DelayedContextFrom(Photo, connection) as cntx:
            for i in pks:
                cntx.bulk_update({
                    Photo._meta.pk.name: i,
                    'camera_model': photo_to_camera_model_mapping.get(i, 'Not set'),
                    'rating': photo_to_rating_mapping.get(i, 0)
                })
        print("Took %s seconds to update 999 photos using Massiviu." % (time.time() - start))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(test_model_context)
    unittest.TextTestRunner(verbosity=2).run(suite)
    suite = unittest.TestLoader().loadTestsFromTestCase(testMassiviu)
    unittest.TextTestRunner(verbosity=2).run(suite)
    suite = unittest.TestLoader().loadTestsFromTestCase(PerformanceTestMassiviu)
    unittest.TextTestRunner(verbosity=2).run(suite)



import os


def configure(appName):
    from django.conf import settings, global_settings
    _INSTALLED_APPS = (
        'django.contrib.auth',
        'django.contrib.contenttypes',
        appName,
    )

    the_module = __import__(appName)

    module_folder = os.path.split(the_module.__file__)[0]
    template_folder = os.path.join(module_folder, 'templates').replace(os.sep, '/')

    ROOT_URLCONF = appName + '.urls'

    settings.configure(default_settings=global_settings,
                       INSTALLED_APPS=_INSTALLED_APPS,
                       DATABASES={'default': {'NAME': ':memory:', 'ENGINE': 'django.db.backends.sqlite3'}},
                       ROOT_URLCONF=ROOT_URLCONF,
                       TEMPLATE_DIRS=(
                           template_folder
                       ),
                       SETTINGS_MODULE=appName
                       )

    from django.core.management import call_command
    import django
    django.setup()

    call_command('makemigrations', appName, interactive=False, verbosity=1)
    call_command('migrate', interactive=False, verbosity=1)



from django.db.models import AutoField


PK_ID = 'id'  # default name of primary key field


class dse_value_parser(classmethod):
    """

    """

    def __init__(self, *args, **kw):
        """

        """
        func = args[0]
        func.is_dse_value_parser = True
        super(dse_value_parser, self).__init__(*args, **kw)


def get_value_parsers_from_class(klass):
    """

    """
    return [getattr(klass, name) for name, parser in klass.__dict__.items() if isinstance(parser, dse_value_parser)]


def get_default_value_for_field_from_model(model, field):
    """
    Get default value, if any, for a specified field in a specified model.
    """
    if not hasattr(model._meta, 'fields'):
        return None

    for f in model._meta.fields:
        if field == f.name:
            if hasattr(f.default, '__name__'):
                if f.default.__name__ == 'NOT_PROVIDED':
                    return None
            return f.default
    return None


def get_fields(table_name, cursor, dbtype, _quote):
    """

    """
    default_sql = 'select * from %s LIMIT 1' % _quote(table_name)
    sql = {
        'sqlite': default_sql,
        'mysql': default_sql,
        'postgres': default_sql
    }
    cursor.execute(sql.get(dbtype, 'select * from %s where 1=2' % _quote(table_name)))
    fields = []
    for idx, field_attrs in enumerate(cursor.description):
        fields.append(field_attrs[0])
    return fields


class ModelContext:
    """

    """
    def __init__(self, model, db_context):
        """

        """
        self.model = model
        self.table_name = model._meta.db_table
        self.object_name = model._meta.object_name
        self.value_parsers = get_value_parsers_from_class(model)
        self.pk = model._meta.pk.name
        self.pk_is_auto_field = isinstance(model._meta.pk, AutoField)
        self.fields = get_fields(self.table_name, db_context.cursor, db_context.db_type, db_context.quote)
        self.default_values = {}

        for key in self.fields:
            if key != self.pk:
                self.default_values[key] = get_default_value_for_field_from_model(model, key)


PARAM_TOKEN = '%s'


class DbContext(object):
    """

    """
    def __init__(self, connection, param_token=PARAM_TOKEN):
        self.param_token = param_token
        self.connection = connection
        self.cursor = connection.cursor()
        mod = self.cursor.connection.__class__.__module__.split('.', 1)[0]

        # a mapping of different database adapters/drivers, needed to handle different
        # quotations/escaping of sql fields, see the quote-method.
        DBNAME_MAP = {
            'psycopg2': 'postgres',
            'MySQLdb': 'mysql',
            'sqlite3': 'sqlite',
            'sqlite': 'sqlite',
            'pysqlite2': 'sqlite'
        }
        self.db_type = DBNAME_MAP.get(mod)
        if self.db_type == 'postgres':
            self.quote = lambda x: '"%s"' % x
        else:
            self.quote = lambda x: '`%s`' % x




class ValueValidatorException(Exception):
    pass


class BaseException(Exception):
    """

    """
    action = 'unknown'

    def __init__(self, msg):
        """

        """
        self.msg = msg

class InsertManyException(BaseException):
    """

    """
    action = 'Insert'

    def __init__(self, exception, table, sql, params):
        self.table = table
        self.sql = sql
        self.params = params
        self.exception = exception

    def __str__(self):
        return "BaseException.%sError on table %s.\nSQL: %s.\nNumber of params: %s.\nException: %s" % \
               (self.action, self.table, self.sql, len(self.params), self.exception)


class UpdateManyException(BaseException):
    """

    """
    action = 'UpdateMany'


class UpdateOneException(BaseException):
    """

    """
    action = 'UpdateOne'

    def __str__(self):
        return "BaseException.%sError on table %s.\nSQL: %s.\nParams: %s.\nException: %s" % \
               (self.action, self.table, '\n'.join(self.sql), self.params, self.exception)


class DeleteManyException(BaseException):
    """

    """
    action = 'DeleteMany'


class PrimaryKeyMissingInInsertValues(BaseException):
    """

    """
    pass


class PrimaryKeyInInsertValues(BaseException):
    """

    """
    pass


from django.db import connection as default_connection
from item_cache import ItemCache
from operations import Operations
from db_context import DbContext
from model_context import ModelContext
from exceptions import ValueValidatorException


class NullLogger:

    def debug(self, msg):
        pass
    def info(self, msg):
        pass
    def warn(self, msg):
        pass
    def warning(self, msg):
        pass
    def error(self, msg):
        pass


class DelayedContextFrom(object):
    """

    """
    def __init__(self, model, connection=default_connection, value_validator_cls=None, logger=NullLogger(), db_context_cls=DbContext, model_context_cls=ModelContext,
                 operations_cls=Operations, item_cache_cls=ItemCache):
        """

        """
        self.logger = logger
        self.db_context = db_context_cls(connection)
        self.model_context = model_context_cls(model, self.db_context)
        self.operations = operations_cls(self.logger, self.db_context, self.model_context)
        self.item_cache = item_cache_cls(self.model_context, self.operations)
        if value_validator_cls and not callable(value_validator_cls):
            raise ValueValidatorException('Supplied value validator is not callable.')
        self.value_validator_cls = value_validator_cls

    def __exit__(self, type, value, traceback):
        """
        Calls flush when exiting the with-block.
        """
        self.item_cache.flush()

    def __enter__(self):
        """
        When using with DelayedContextFrom(...) as cntx:
        """
        return self

    def validate_values_with(self, value_validator_cls):
        """

        :param value_validator_cls:
        :return:
        """
        self.value_validator_cls = value_validator_cls
        return self

    def reset(self):
        """

        """
        self.item_cache.reset()

    def update(self, values):
        """
        Adds a set of values to execute as update using cursor.executemany.
        """
        if self.value_validator_cls:
            values = self.value_validator_cls(values)

        self.item_cache.update(values)

    def bulk_update(self, values):
        """
        Adds a set of values to use for alternative bulk updates, using Model.objects.filter(id__in=...).update().
        See the README.md for more details.
        """
        if self.value_validator_cls:
            values = self.value_validator_cls(values)

        self.item_cache.bulk_update(values)

    def insert(self, values):
        """
        Adds a dictionary with values to insert/update
        """
        if self.value_validator_cls:
            values = self.value_validator_cls(values)

        self.item_cache.insert(values)

    def delete(self, pk):
        """

        """
        self.item_cache.delete(pk)




#!/usr/bin/env python

try:
    from setuptools import setup
except (ImportError):
    from distribute import setup

setup(name='massiviu',
      version='0.1.0',
      description='MassivIU - Simplified "bulk" insert/update/delete for Django.',
      author='Thomas Weholt',
      author_email='thomas@weholt.org',
      long_description=open('README.md').read(),
      include_package_data=True,
      packages=['massiviu', ],
      url="https://github.com/weholt/massiviu",
      requires=['django'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Framework :: Django',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Database',
      ],
)




from exceptions import InsertManyException, PrimaryKeyMissingInInsertValues, \
    PrimaryKeyInInsertValues, UpdateManyException, UpdateOneException


class Operations(object):
    """
    
    """
    def __init__(self, logger, db_context, model_context, debug=True):
        """

        """
        self.debug = debug
        self.db_context = db_context
        self.model_context = model_context
        self.logger = logger
        self.insert_sql = self._generate_insert_sql()
        self.update_sql = self._generate_update_sql()
        self.records_processed = 0
        self.sql_calls = 0
        self.do_clean_house = True

    def _generate_insert_sql(self):
        sql = 'insert into %s (%s) values (%s)' % (
            self.db_context.quote(self.model_context.table_name),
            ','.join(self.db_context.quote(f) for f in self.model_context.fields if f != self.model_context.pk or not self.model_context.pk_is_auto_field),
            ','.join(self.db_context.param_token for f in self.model_context.fields if f != self.model_context.pk or not self.model_context.pk_is_auto_field),
        )
        return sql

    def _generate_update_sql(self):
        sql = ['update %s set' % self.db_context.quote(self.model_context.table_name)]
        m = []
        for field_name in self.model_context.fields:
            if field_name == self.model_context.pk:
                continue
            m.append("%s = %s" % (self.db_context.quote(field_name), self.db_context.param_token))
        sql.append(',\n'.join(m))
        sql.append('where %s = %s' % (self.db_context.quote(self.model_context.pk), self.db_context.param_token))
        return '\n'.join(sql)

    def execute_insert_statements(self, insert_items):
        """
        Executes all bulk insert statements.
        """
        field_values = []
        for items in insert_items:
            m = []
            for field_name in self.model_context.fields:
                if field_name in items:
                    m.append(items[field_name])
                elif field_name != self.model_context.pk or not self.model_context.pk_is_auto_field:
                    m.append(None)
            field_values.append(m)
            self.records_processed += 1

        if self.debug:
            self.logger.debug("Executing insert: %s" % self.insert_sql)
            for f in field_values:
                self.logger.debug(str(f))
        try:
            self._execute(self.insert_sql, field_values, many=True)
        except Exception as e:
            raise InsertManyException(e, self.model_context.table_name, self.insert_sql, field_values)

    def execute_delete_statements(self, delete_items):
        """
        Executes all bulk delete statements.
        """
        self.model_context.model.objects.filter(**{"%s__in" % self.model_context.pk: delete_items}).delete()
        self.records_processed += 1

    def execute_sql(self, item_cache):
        """
        Executes all cached sql statements.
        """
        if item_cache.bulk_updates:
            self.execute_bulk_updates(item_cache.bulk_updates)

        if item_cache.update_items:
            self.execute_updates(item_cache.update_items)

        if item_cache.insert_items:
            self.execute_insert_statements(item_cache.insert_items)

        if item_cache.delete_items:
            self.execute_delete_statements(item_cache.delete_items)

    def _execute(self, sql, field_values, many=True):
        self.sql_calls += 1
        try:
            if many:
                self.cursor.executemany(sql, field_values)
            else:
                self.cursor.execute(sql, field_values)
        except:
            self.cursor = self.db_context.cursor
            if many:
                self.cursor.executemany(sql, field_values)
            else:
                self.cursor.execute(sql, field_values)
        finally:
            self.clean_house()

    def execute_updates(self, update_items):
        """
        Executes all bulk update statements.
        """
        params_for_executemany = []
        params_for_execute = []
        # If there all fields are present we can optimize and use executemany,
        # if not we must execute each SQL call in sequence
        for items in update_items:
            if len(items.keys()) != len(self.model_context.fields):
                params_for_execute.append(items)
            else:
                found_all_fields = True
                for field in self.model_context.fields:
                    if not field in items:
                        found_all_fields = False
                        break

                if found_all_fields:
                    params_for_executemany.append(items)
                else:
                    params_for_execute.append(items)

        if params_for_executemany:
            field_values = []
            for items in params_for_executemany:
                m = []
                for field_name in self.model_context.fields:
                    if field_name == self.model_context.pk:
                        continue
                    if field_name in items:
                        m.append(items[field_name])
                    else:
                        m.append(None)
                m.append(items.get(self.model_context.pk))
                field_values.append(m)
                self.records_processed += 1

            if self.debug:
                self.logger.debug("Executing update: %s" % self.update_sql)
                for f in field_values:
                    self.logger.debug(str(f))

            try:
                self._execute(self.update_sql, field_values, many=True)
            except Exception as e:
                raise UpdateManyException(e, self.model_context.table_name, self.update_sql, field_values)

        for items in params_for_execute:
            sql = ['update %s set' % self.db_context.quote(self.model_context.table_name)]
            m = []
            field_values = []
            for field_name in items.keys():
                if field_name == self.model_context.pk or field_name not in self.model_context.fields:
                    continue
                m.append("%s = %s" % (self.db_context.quote(field_name), self.db_context.param_token))
                field_values.append(items[field_name])
            sql.append(',\n'.join(m))
            sql.append('where %s = %s' % (self.db_context.quote(self.model_context.pk), self.db_context.param_token))
            field_values.append(items[self.model_context.pk])
            self.records_processed += 1
            if self.debug:
                self.logger.debug("Executing update: %s" % '\n'.join(sql))
                for f in field_values:
                    self.logger.debug(str(f))

            try:
                self._execute('\n'.join(sql), field_values, many=False)
            except Exception as e:
                raise UpdateOneException(e, self.model_context.table_name, sql, field_values)

    def execute_bulk_updates(self, bulk_updates):
        """
        Executes the bulk bulk update statements
        """
        for field, values in bulk_updates.items():
            for value, ids in values.items():
                self.model_context.model.objects.filter(**{"%s__in" % self.model_context.pk: ids}).update(**{field: value})
                self.records_processed += 1

    def clean_house(self):
        """
        This method removes the last query from the list of queries stored in the django connection
        object. The django-debug-toolbar modifies that list and if we leave our dse based query lying around
        it will cause the debug-toolbar to crash.

        To disable this feature set dse.CLEAN_HOUSE = False.
        This method might later on be used for other things as well.
        """
        #if self.do_clean_house:
         #   self.db_context.connection.queries = self.db_context.connection.queries[:-1]
        pass

